from __future__ import annotations

import numpy as np
import pandas as pd
from pandas import DatetimeIndex
from pandas.core.series import Series


def since_event(event: Series[bool]) -> Series[float]:
    """Calculate the elapsed time since the last event.

    Parameters
    ----------
    event : Series[bool]
        Whether the event occurred.

    Returns
    -------
    Series[float]
        The difference between the index of the last event
        and the current index.
        If the index is a DatetimeIndex, the unit is hours.
    """
    event = event.astype(bool, copy=True)
    orig_idx = event.index
    idx = event.index
    if isinstance(idx, DatetimeIndex):
        # without converting to float, calculation will be extremely slow
        idx = pd.to_numeric(idx, downcast="float") / 1e9 / 60 / 60
        event.index = idx
    last_event_index = Series(np.nan, index=idx)
    last_event_index[event] = idx
    last_event_index.ffill(inplace=True)
    res = idx - last_event_index
    res.index = orig_idx
    return res


def until_event(event: Series[bool]) -> Series[float]:
    """Calculate the elapsed time until the next event.

    Parameters
    ----------
    event : Series[bool]
        Whether the event occurred.

    Returns
    -------
    Series[float]
        The difference between the index of the next event
        and the current index.
        If the index is a DatetimeIndex, the unit is hours.
    """
    event = event.astype(bool, copy=True)
    orig_idx = event.index
    idx = event.index
    if isinstance(event.index, DatetimeIndex):
        # without converting to float, calculation will be extremely slow
        idx = pd.to_numeric(idx, downcast="float") / 1e9 / 60 / 60
        event.index = idx
    next_event_index = Series(np.nan, index=idx)
    next_event_index[event] = idx
    next_event_index.bfill(inplace=True)
    res = next_event_index - idx
    res.index = orig_idx
    return res



from __future__ import annotations

from typing import Any, Generic, Mapping, Sequence

import numpy as np
from joblib import Parallel, delayed
from numpy.typing import NDArray
from typing_extensions import Self

from .estimator_wrapper import EstimatorWrapperBase
from .types import TEstimator


class RecursiveFitSubtractRegressor(
    Generic[TEstimator], EstimatorWrapperBase[TEstimator]
):
    """Regressor that fits the residual of the prediction of the previous model."""

    estimators: Sequence[TEstimator]

    def __init__(
        self, estimators: Sequence[TEstimator], *, n_jobs: int | None = None
    ) -> None:
        """Regressor that fits the residual of the prediction of the previous model.

        Parameters
        ----------
        estimators : Sequence[TEstimator]
            The estimators to be wrapped.
        n_jobs : int | None, optional
            The number of jobs to run in parallel when predicting, by default None
            Note that fitting cannot be parallelized.
        """
        self.estimators = estimators
        self.n_jobs = n_jobs

    def fit(
        self,
        X: Any,
        y: Any,
        predict_params: Mapping[str, Any] | None = None,
        **fit_params: Any,
    ) -> Self:
        if predict_params is None:
            predict_params = {}
        for i, model in enumerate(self.estimators):
            model.fit(X, y, **fit_params)
            if i < len(self.estimators) - 1:
                y -= model.predict(X, **predict_params)
        return self

    def predict(self, X: Any, **predict_params: Any) -> NDArray[Any]:
        parrarel_results = Parallel(n_jobs=self.n_jobs)(
            delayed(model.predict)(X, **predict_params) for model in self.estimators
        )
        if parrarel_results is None:
            raise ValueError("joblib.Parallel returned None")
        return np.sum(parrarel_results, axis=0)

    def predict_var(self, X: Any, **predict_params: Any) -> NDArray[Any]:
        return self.estimator.predict_var(X, **predict_params)

    @property
    def estimator(self) -> TEstimator:  # type: ignore
        return self.estimators[-1]



__version__ = "0.0.0"
from .append_prediction_to_x import AppendPredictionToX, AppendPredictionToXSingle
from .append_x_prediction_to_x import AppendXPredictionToX
from .drop_by_noise_prediction import DropByNoisePrediction
from .drop_missing_columns import DropMissingColumns
from .drop_missing_rows_y import DropMissingRowsY
from .estimator_wrapper import EstimatorWrapperBase
from .eval_set import EvalSetWrapper
from .event import since_event, until_event
from .id_transformer import IdTransformer
from .intersect import IntersectXY
from .pandas import (
    DataFrameWrapper,
    ExcludedColumnTransformerPandas,
    FeatureUnionPandas,
    IncludedColumnTransformerPandas,
    SmartMultioutputEstimator,
)
from .proba import (
    ComposeVarEstimator,
    DummyRegressorVar,
    PipelineVar,
    StandardScalerVar,
    TransformedTargetEstimatorVar,
)
from .recursive_fit_subtract_regressor import RecursiveFitSubtractRegressor
from .reindex_missing_columns import ReindexMissingColumns
from .report_non_finite import ReportNonFinite

__all__ = [
    "DataFrameWrapper",
    "FeatureUnionPandas",
    "IncludedColumnTransformerPandas",
    "ExcludedColumnTransformerPandas",
    "AppendPredictionToX",
    "AppendPredictionToXSingle",
    "AppendXPredictionToX",
    "DropByNoisePrediction",
    "DropMissingColumns",
    "DropMissingRowsY",
    "EstimatorWrapperBase",
    "until_event",
    "since_event",
    "IdTransformer",
    "RecursiveFitSubtractRegressor",
    "IntersectXY",
    "SmartMultioutputEstimator",
    "ComposeVarEstimator",
    "DummyRegressorVar",
    "TransformedTargetEstimatorVar",
    "PipelineVar",
    "StandardScalerVar",
    "EvalSetWrapper",
    "ReportNonFinite",
    "ReindexMissingColumns",
]



from __future__ import annotations

from numbers import Real
from typing import Any, Literal

import numpy as np
from lightgbm import LGBMRegressor
from numpy.random import RandomState
from numpy.typing import NDArray
from sklearn.base import clone
from sklearn.feature_selection import SelectFromModel
from sklearn.utils import check_random_state
from sklearn.utils._param_validation import Interval, StrOptions
from typing_extensions import Self


class DropByNoisePrediction(SelectFromModel):
    """Remove features based on their importance to a model's prediction of noise.

    "Unsupervised Learning by Predicting Noise"
    https://arxiv.org/pdf/1704.05310.pdf
    https://ar5iv.labs.arxiv.org/html/1704.05310

    "Neural Architecture Search with Random Labels"
    https://arxiv.org/abs/2101.11834
    https://ar5iv.labs.arxiv.org/html/2101.11834

    Original Implementation:
    https://gist.github.com/richmanbtc/075178cd0e6d15c4a251128068991d47
    """

    def __init__(
        self,
        estimator: Any | None = None,
        *,
        drop_rate: float = 0.1,
        distribution: Literal["uniform", "normal", "arange"] = "uniform",
        random_state: RandomState | int | None = None,
    ) -> None:
        """Remove features based on their importance to a model's prediction of noise.

        "Unsupervised Learning by Predicting Noise"
        https://arxiv.org/pdf/1704.05310.pdf
        https://ar5iv.labs.arxiv.org/html/1704.05310

        "Neural Architecture Search with Random Labels"
        https://arxiv.org/abs/2101.11834
        https://ar5iv.labs.arxiv.org/html/2101.11834

        Original Implementation:
        https://gist.github.com/richmanbtc/075178cd0e6d15c4a251128068991d47

        Parameters
        ----------
        estimator : Any, optional
            Estimator to use for feature importance.
            If None, uses a default LGBMestimator,
            by default None
        percentile : float, optional
            Percent of features to keep, by default 10
        distribution : Literal['uniform', 'normal', 'arange'], optional
            Distribution to use for the target, by default &quot;uniform&quot;
        """
        self._parameter_constraints.update(
            {
                "drop_rate": [Interval(Real, 0, 1, closed="both")],
                "distribution": [StrOptions({"uniform", "normal", "arange"})],
            }
        )
        estimator = (
            LGBMRegressor(n_jobs=-1, random_state=random_state)
            if estimator is None
            else clone(estimator)
        )
        self.drop_rate = drop_rate
        self.distribution = distribution
        self.random_state = random_state
        check_random_state(self.random_state)
        super().__init__(
            estimator=estimator,
            threshold=-np.inf,
            prefit=False,
            importance_getter=lambda x: -x.feature_importances_,
            max_features=lambda x: int(x.shape[1] * (1 - drop_rate)),
        )

    def _generate_y(self, X: NDArray[Any]) -> NDArray[Any]:
        self.random_state_ = check_random_state(self.random_state)
        if self.distribution == "arange":
            y = np.arange(X.shape[0])
        elif hasattr(self.random_state_, self.distribution):
            y = getattr(self.random_state_, self.distribution)(0, 1, size=X.shape[0])
        else:
            raise ValueError(f"Invalid distribution: {self.distribution}")
        return y

    def fit(self, X: Any, y: Any = None, **fit_params: Any) -> Self:
        y = self._generate_y(X)
        return super().fit(X, y, **fit_params)

    def fit_transform(self, X: Any, y: Any = None, **fit_params: Any) -> Any:
        y = self._generate_y(X)
        return super().fit_transform(X, y, **fit_params)



from .compose_var import ComposeVarEstimator
from .dummy_regressor import DummyRegressorVar
from .pipeline_var import PipelineVar
from .standard_scaler_var import StandardScalerVar
from .transformed_target_estimator import TransformedTargetEstimatorVar

__all__ = [
    "ComposeVarEstimator",
    "DummyRegressorVar",
    "PipelineVar",
    "TransformedTargetEstimatorVar",
    "StandardScalerVar",
]



from __future__ import annotations

from typing import Any, Callable

import torch
import torch.nn as nn
from sklearn.base import BaseEstimator, TransformerMixin, check_is_fitted
from typing_extensions import Self


def svd_flip(
    u: torch.Tensor, v: torch.Tensor, u_based_decision: bool = True
) -> tuple[torch.Tensor, torch.Tensor]:
    """Sign correction to ensure deterministic output from SVD.

    Adjusts the columns of u and the rows of v such that the loadings in the
    columns in u that are largest in absolute value are always positive.

    Parameters
    ----------
    u : torch.Tensor
        Parameters u and v are the output of `linalg.svd` or
        :func:`~sklearn.utils.extmath.randomized_svd`, with matching inner
        dimensions so one can compute `np.dot(u * s, v)`.

    v : torch.Tensor
        Parameters u and v are the output of `linalg.svd` or
        :func:`~sklearn.utils.extmath.randomized_svd`, with matching inner
        dimensions so one can compute `np.dot(u * s, v)`.
        The input v should really be called vt to be consistent with scipy's
        output.

    u_based_decision : bool, default=True
        If True, use the columns of u as the basis for sign flipping.
        Otherwise, use the rows of v. The choice of which variable to base the
        decision on is generally algorithm dependent.

    Returns
    -------
    u_adjusted : torch.Tensor
        Array u with adjusted columns and the same dimensions as u.

    v_adjusted : torch.Tensor
        Array v with adjusted rows and the same dimensions as v.
    """
    if u_based_decision:
        max_abs_cols = torch.argmax(torch.abs(u), dim=0)
        signs = torch.sign(u[max_abs_cols, torch.arange(u.shape[1])])
    else:
        max_abs_rows = torch.argmax(torch.abs(v), dim=1)
        signs = torch.sign(v[torch.arange(v.shape[0]), max_abs_rows])
    u *= signs
    v *= signs[:, None]
    return u, v


def wrap_torch(func: Callable[..., Any]) -> Callable[..., Any]:
    """Wrap a function to convert all non-torch.Tensor arguments to torch.Tensor
    and convert the result to numpy.ndarray.

    Parameters
    ----------
    func : Callable[..., Any]
        The function to wrap.

    Returns
    -------
    Callable[..., Any]
        The wrapped function.
    """

    def wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
        args_ = list(args)
        kwargs_ = dict(kwargs)
        for i, arg in enumerate(args_):
            if isinstance(arg, torch.Tensor):
                continue
            try:
                args_[i] = torch.as_tensor(arg, device=self.device, dtype=self.dtype)
            except TypeError:
                pass
        for k, v in kwargs_.items():
            if isinstance(v, torch.Tensor):
                continue
            try:
                kwargs_[k] = torch.as_tensor(v, device=self.device, dtype=self.dtype)
            except TypeError:
                pass

        result = func(self, *args_, **kwargs_)
        if isinstance(result, torch.Tensor):
            return result.detach().cpu().numpy()
        return result

    return wrapper


class PCATorch(nn.Module, BaseEstimator, TransformerMixin):
    """PCA using torch.linalg.svd.

    If using CUDA, the first call may take significantly long time (~2s)
    due to CUDA initialization, but the subsequent calls should be faster
    than sklearn.decomposition.PCA, although the algorithm might be less efficient.

    Call `python -m sklearn_utilities.torch.pca 10000x100` to test the performance.

    If we could easily replace `np` with `torch` in sklearn...

    Attributes
    ----------
    mean_ : torch.Tensor
        The mean vector.
    components_ : torch.Tensor
        `Vt` where
        `X = U D Vt`
        `Y = X Vh`"""

    def __init__(
        self,
        n_components: int | None = None,
        *,
        qr: bool = False,
        svd_flip: bool | None = None,
        device: torch.device | int | str = "cuda"
        if torch.cuda.is_available()
        else "cpu",
        dtype: torch.dtype = torch.float32,
        **kwargs: Any,
    ) -> None:
        """PCA using torch.linalg.svd.
        Might be faster than sklearn.decomposition.PCA
        if using GPU, but the algorithm is less efficient.

        If we could easily replace `np` with `torch` in sklearn...

        Parameters
        ----------
        n_components : int | None, optional
            Number of components to keep, by default None
        qr : bool, optional
            Whether to use QR decomposition, by default False
        svd_flip : bool | None, optional
            Whether to flip the sign of the components, by default None
            If None, the sign will be flipped if `qr` is False
            If svd_flip is not used, the results might not be consistent
            with sklearn.decomposition.PCA
        device : torch.device | int | str, optional
            The device to use, by default
            `"cuda" if torch.cuda.is_available() else "cpu"`
        dtype : torch.dtype, optional
            The dtype to use, by default torch.float32

        Attributes
        ----------
        mean_ : torch.Tensor
            The mean vector.
        components_ : torch.Tensor
            `Vt` where
            `X = U D Vt`
            `Y = X Vh`"""
        super().__init__()
        self.n_components = n_components
        self.qr = qr
        self.svd_flip = svd_flip
        self.device = device
        self.dtype = dtype
        self.kwargs = kwargs

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        return self.transform(X)

    @wrap_torch
    @torch.no_grad()
    def fit(self, X: torch.Tensor) -> Self:
        _, d = X.size()
        if self.n_components is not None:
            d = min(self.n_components, d)
        self.register_buffer("mean_", X.mean(0, keepdim=True))
        Xc = X - self.mean_
        to_svd_flip = self.svd_flip
        if to_svd_flip is None:
            to_svd_flip = not self.qr
        if self.qr:
            Q, R = torch.linalg.qr(Xc)
            U, _, Vt = torch.linalg.svd(R, full_matrices=False)
            if to_svd_flip:
                U, Vt = svd_flip(Q @ U, Vt)
        else:
            U, _, Vt = torch.linalg.svd(Xc, full_matrices=False)
            if to_svd_flip:
                U, Vt = svd_flip(
                    U, Vt
                )  # to be deterministic and consistent with sklearn
        self.register_buffer("components_", Vt[:d])
        return self

    @wrap_torch
    def transform(self, X: torch.Tensor) -> torch.Tensor:
        check_is_fitted(self, ["mean_", "components_"])
        Xc = X - self.mean_
        return Xc @ self.components_.T

    @wrap_torch
    def inverse_transform(self, X: torch.Tensor) -> torch.Tensor:
        check_is_fitted(self, ["mean_", "components_"])
        return (X @ self.components_) + self.mean_


def pca_performance_test() -> None:
    import sys
    from time import perf_counter

    import torch
    from sklearn.datasets import make_regression
    from sklearn.decomposition import PCA

    if len(sys.argv) < 2:
        size = (1000, 1000)
    else:
        size = tuple(int(length) for length in sys.argv[1].split("x"))  # type: ignore

    if not torch.cuda.is_available():
        raise RuntimeError("CUDA is not available")

    torch.set_num_threads(32)
    torch.backends.cudnn.benchmark = True
    torch.set_float32_matmul_precision("medium")

    start = perf_counter()
    a = torch.randn(1000, 1000, device="cuda")
    b = torch.randn(1000, 1000, device="cuda")
    torch.matmul(a, b)
    elapsed = perf_counter() - start

    print(f"Warming up torch: {elapsed:g}[s]")

    elapsed_dict = {}

    X, _ = make_regression(n_samples=size[0], n_features=size[1])

    pcas = {
        "sklearn": PCA(n_components=2),
        "torch-cuda": PCATorch(n_components=2, qr=False, svd_flip=False, device="cuda"),
        "torch-cuda-svd_flip": PCATorch(
            n_components=2, qr=False, svd_flip=True, device="cuda"
        ),
        "torch-cuda-qr": PCATorch(
            n_components=2, qr=True, svd_flip=False, device="cuda"
        ),
        "torch-cuda-qr-svd_flip": PCATorch(
            n_components=2, qr=True, svd_flip=True, device="cuda"
        ),
        "torch-cpu": PCATorch(n_components=2, qr=False, svd_flip=False, device="cpu"),
        "torch-cpu-svd_flip": PCATorch(
            n_components=2, qr=False, svd_flip=True, device="cpu"
        ),
        "torch-cpu-qr": PCATorch(n_components=2, qr=True, svd_flip=False, device="cpu"),
        "torch-cpu-qr-svd_flip": PCATorch(
            n_components=2, qr=True, svd_flip=True, device="cpu"
        ),
    }

    for name, pca in pcas.items():
        for i in range(2):
            start = perf_counter()
            pca.fit(X)
            pca.transform(X)
            elapsed = perf_counter() - start
            print(f"{name}: {elapsed:g}[s]")
        elapsed_dict[name] = elapsed

    print(elapsed_dict)


if __name__ == "__main__":
    pca_performance_test()



from .proba import (
    AlgebraicErrors,
    AllowNan,
    AsymmetricLoss,
    AsymmetricLosses,
    LNErrors,
    LogCoshErrors,
    SkorchCNNReshaperProba,
    SkorchReshaperProba,
    XSigmoidErrors,
    XTanhErrors,
)
from .reshaper import SkorchCNNReshaper, SkorchReshaper

__all__ = [
    "SkorchReshaper",
    "SkorchCNNReshaper",
    "SkorchReshaperProba",
    "SkorchCNNReshaperProba",
    "LNErrors",
    "LogCoshErrors",
    "XSigmoidErrors",
    "XTanhErrors",
    "AlgebraicErrors",
    "AllowNan",
    "AsymmetricLoss",
    "AsymmetricLosses",
    "SkorchReshaperProba",
    "SkorchCNNReshaperProba",
]



from __future__ import annotations

from typing import Any, Generic

from sklearn.base import BaseEstimator, MetaEstimatorMixin

from .types import TEstimator


class EstimatorWrapperBase(BaseEstimator, MetaEstimatorMixin, Generic[TEstimator]):
    """A base class for estimator wrappers
    that delegates all attributes to the wrapped estimator."""

    estimator: TEstimator

    def __init__(self, estimator: TEstimator) -> None:
        """A base class for estimator wrappers
        that delegates all attributes to the wrapped estimator.

        Parameters
        ----------
        estimator : Any
            The estimator to be wrapped.
        """
        self.estimator = estimator

    def __getattribute__(self, __name: str) -> Any:
        try:
            return object.__getattribute__(self, __name)
        except AttributeError:
            return getattr(self.estimator, __name)

    def __setattr__(self, __name: str, __value: Any) -> None:
        try:
            return object.__setattr__(self, __name, __value)
        except AttributeError:
            return setattr(self.estimator, __name, __value)

    def __delattr__(self, __name: str) -> None:
        try:
            return object.__delattr__(self, __name)
        except AttributeError:
            return delattr(self.estimator, __name)

    def __getitem__(self, __key: str) -> Any:
        return self.estimator.__getitem__(__key)

    # Due to the bug in Python, __instancecheck__ does not work.
    # https://bugs.python.org/issue35083

    # This makes the class sklearn.clone() incompatible.
    """@property
    def __class__(self) -> Any:
        return self.estimator.__class__

    @__class__.setter
    def __class__(self, __class__: Any) -> None:  # noqa
        self.estimator.__class__ = __class__"""



from typing import Any

from sklearn.pipeline import Pipeline


class PipelineVar(Pipeline):
    """Pipeline that supports predict_var method"""

    def predict_var(self, X: Any, **predict_params: Any) -> Any:
        Xt = X
        for _, name, transform in self._iter(with_final=False):
            Xt = transform.transform(Xt)
        return self.steps[-1][-1].predict_var(Xt, **predict_params)



from __future__ import annotations

from logging import getLogger
from pathlib import Path
from typing import Any, Callable

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.figure import Figure
from pandas import DataFrame, Series, Timestamp
from sklearn.base import BaseEstimator, TransformerMixin
from typing_extensions import Self

from .types import TXPandas

LOG = getLogger(__name__)


class ReportNonFinite(BaseEstimator, TransformerMixin):
    """Report non-finite values in X or y."""

    def __init__(
        self,
        *,
        on_fit: bool = False,
        on_fit_y: bool = False,
        on_transform: bool = True,
        plot: bool = True,
        calc_corr: bool = False,
        callback: Callable[[dict[str, DataFrame | Series]], None] | None = None,
        callback_figure: Callable[[Figure], None]
        | None = lambda fig: Path("sklearn_utilities_info/ReportNonFinite").mkdir(  # type: ignore
            parents=True, exist_ok=True
        )
        or fig.savefig(
            Path("sklearn_utilities_info/ReportNonFinite")
            / f"{Timestamp.now().isoformat().replace(':', '-')}.png"
        ),
    ) -> None:
        """Report non-finite values in X or y.

        Parameters
        ----------
        on_fit : bool, optional
            Whether to report non-finite values in X during fit, by default False
        on_fit_y : bool, optional
            Whether to report non-finite values in y during fit, by default False
        on_transform : bool, optional
            Whether to report non-finite values in X during transform, by default True
        plot : bool, optional
            Whether to plot the report result, by default True
        calc_corr : bool, optional
            Whether to calculate the correlation of non-finite values, by default False
        callback : Callable[[dict[str, DataFrame  |  Series]], None] | None, optional
            The callback function, by default None
        callback_figure : _type_, optional
            The callback function for figure, by default
            `lambda fig:
            Path("sklearn-utilities/ReportNonFinite").mkdir(parents=True, exist_ok=True)
            or fig.savefig(Path("sklearn-utilities/ReportNonFinite") /
            f"{Timestamp.now().isoformat().replace(':', '-')}.png")`
        """
        self.on_fit = on_fit
        self.on_fit_y = on_fit_y
        self.on_transform = on_transform
        self.plot = plot
        self.calc_corr = calc_corr
        self.callback = callback
        self.callback_figure = callback_figure

    def fit(self, X: DataFrame, y: Any = None, **fit_params: Any) -> Self:
        if self.on_fit:
            try:
                self._report(X, "fit")
            except Exception as e:
                LOG.warning(f"Failed to report non-finite values in X during fit: {e}")
                LOG.exception(e)

        if self.on_fit_y:
            try:
                DataFrame(y)
            except Exception as e:
                LOG.warning(f"Failed to convert y to DataFrame during fit: {e}")
                LOG.exception(e)

            try:
                self._report(DataFrame(y), "fit_y")
            except Exception as e:
                LOG.warning(f"Failed to report non-finite values in y during fit: {e}")
                LOG.exception(e)
        return self

    def transform(self, X: TXPandas, y: Any = None, **fit_params: Any) -> TXPandas:
        if self.on_transform:
            try:
                self._report(X, "transform")
            except Exception as e:
                LOG.warning(
                    f"Failed to report non-finite values in X during transform: {e}"
                )
                LOG.exception(e)
        return X

    def _report(self, X: TXPandas, caller: str = "") -> TXPandas:
        """Report non-finite values in X.

        Parameters
        ----------
        X : TXPandas
            Input data.
        caller : str, optional
            The caller name used in the log message, by default "".

        Returns
        -------
        TXPandas
            Input data.
        """
        is_na = X.isna()
        is_inf = X.isin([np.inf, -np.inf])
        is_non_finite = is_na | is_inf

        d: dict[str, DataFrame | Series] = {
            "nan_rate_by_column": is_na.mean(),
            "inf_rate_by_column": is_inf.mean(),
            "nan_rate_by_row": is_na.mean(axis=1),
            "inf_rate_by_row": is_inf.mean(axis=1),
        }
        d = d | {
            "non_finite_rate_by_column": d["nan_rate_by_column"]
            + d["inf_rate_by_column"],
            "non_finite_rate_by_row": d["nan_rate_by_row"] + d["inf_rate_by_row"],
        }

        if self.calc_corr:
            d["nan_rate_corr_by_column"] = is_na.corr()
            d["inf_rate_corr_by_column"] = is_inf.corr()
            d["non_finite_rate_corr_by_column"] = is_non_finite.corr()

        LOG.info(f"Non-finite values in X during {caller}: {d}")

        if self.plot:
            import seaborn as sns

            fig, axes = plt.subplots(3, 3 if self.calc_corr else 2, figsize=(20, 10))
            fig.suptitle(f"Non-finite values in X during {caller}")
            d["nan_rate_by_column"].plot(
                ax=axes[0, 0],
                kind="bar",
                title="NaN rate By column",
                xlabel="column name",
                ylabel="NaN rate",
            )
            d["inf_rate_by_column"].plot(
                ax=axes[1, 0],
                kind="bar",
                title="Inf rate By column",
                xlabel="column name",
                ylabel="Inf rate",
            )
            d["non_finite_rate_by_column"].plot(
                ax=axes[2, 0],
                kind="bar",
                title="Non-finite rate By column",
                xlabel="column name",
                ylabel="Non-finite rate",
            )
            d["nan_rate_by_row"].plot(
                ax=axes[0, 1],
                kind="line",
                title="NaN rate By row",
                xlabel="row index",
                ylabel="NaN rate",
            )
            d["inf_rate_by_row"].plot(
                ax=axes[1, 1],
                kind="line",
                title="Inf rate By row",
                xlabel="row index",
                ylabel="Inf rate",
            )
            d["non_finite_rate_by_row"].plot(
                ax=axes[2, 1],
                kind="line",
                title="Non-finite rate By row",
                xlabel="row index",
                ylabel="Non-finite rate",
            )
            if self.calc_corr:
                sns.heatmap(
                    d["nan_rate_corr_by_column"], ax=axes[0, 2], vmin=-1, vmax=1
                )
                axes[0, 2].set_title("NaN rate Corr By column")
                sns.heatmap(
                    d["inf_rate_corr_by_column"], ax=axes[1, 2], vmin=-1, vmax=1
                )
                axes[1, 2].set_title("Inf rate Corr By column")
                sns.heatmap(
                    d["non_finite_rate_corr_by_column"], ax=axes[2, 2], vmin=-1, vmax=1
                )
                axes[2, 2].set_title("Non-finite rate Corr By column")

            # tight layout
            plt.tight_layout()

            # callback
            if self.callback_figure is not None:
                self.callback_figure(fig)

        if self.callback is not None:
            self.callback(d)
        return X



from .column_transformer_pandas import (
    ExcludedColumnTransformerPandas,
    IncludedColumnTransformerPandas,
)
from .dataframe_wrapper import DataFrameWrapper
from .feature_union_pandas import FeatureUnionPandas
from .multioutput import SmartMultioutputEstimator

__all__ = [
    "DataFrameWrapper",
    "FeatureUnionPandas",
    "IncludedColumnTransformerPandas",
    "ExcludedColumnTransformerPandas",
    "SmartMultioutputEstimator",
]



from __future__ import annotations

from typing import Any, TypeVar

from sklearn.base import BaseEstimator, TransformerMixin
from typing_extensions import Self

T = TypeVar("T", bound=Any)


class IdTransformer(BaseEstimator, TransformerMixin):
    """A transformer that does nothing."""

    def fit(self, X: Any, y: Any = None, **fit_params: Any) -> Self:
        return self

    def transform(self, X: T, **transform_params: Any) -> T:
        return X

    def inverse_transform(self, X: T, **transform_params: Any) -> T:
        return X

    def inverse_transform_var(self, X: T, **transform_params: Any) -> T:
        return X



from __future__ import annotations

from typing import Any, Generic, Sequence

import numpy as np
from joblib import Parallel, delayed
from pandas import DataFrame, Series, concat
from sklearn.base import BaseEstimator, TransformerMixin, check_is_fitted, clone
from typing_extensions import Self

from .types import TX, TEstimator


def generate_new_prefix(X: DataFrame, prefix: str = "y_pred_") -> str:
    """Generate a new column name for the prediction."""
    last_number = (
        X.columns.to_series()
        .str.extract(rf"^{prefix}(\d+)_.+$", expand=False)
        .fillna(-1)
        .astype(int)
        .max()
    )
    return f"{prefix}{last_number + 1}"


class AppendPredictionToXSingle(BaseEstimator, TransformerMixin, Generic[TEstimator]):
    """Append the prediction of the estimator to X.
    To use multiple estimators, use AppendPredictionToX instead."""

    estimator: TEstimator
    estimator_: TEstimator
    """The fitted estimator."""

    def __init__(self, estimator: TEstimator, *, concat: bool = True) -> None:
        """Append the prediction of the estimator to X.

        If pandas DataFrame is given, the prediction is added as a new column
        with the name "y_pred_{estimator.__class__.__name__}_{i}" if the prediction
        is 1D, or "y_pred_{estimator.__class__.__name__}_{i}_{column_name}" if the
        prediction is 2D.

        To use multiple estimators, use AppendPredictionToX instead.

        Parameters
        ----------
        estimator : Any
            The estimator to be wrapped.
        concat : bool, optional
            Whether to concatenate the prediction to X,
            by default True
        """
        self.estimator = estimator
        self.concat = concat

    def fit(self, X: Any, y: Any = None, **fit_params: Any) -> Self:
        """Fit the estimator."""
        self.estimator_ = clone(self.estimator).fit(X, y, **fit_params)
        return self

    def transform(self, X: TX, y: Any = None, **predict_params: Any) -> TX:
        """Append the prediction of the estimator to X.
        If pandas DataFrame is given, the prediction is added as a new column
        with the name "y_pred_{estimator.__class__.__name__}_{i}" if the prediction
        is 1D, or "y_pred_{estimator.__class__.__name__}_{i}_{column_name}" if the
        prediction is 2D."""
        check_is_fitted(self, "estimator_")
        y_pred = self.estimator_.predict(X, **predict_params)

        # concat the prediction X
        if isinstance(X, DataFrame):
            # pandas

            # add prefix
            prefix = f"y_pred_{self.estimator_.__class__.__name__}_"
            prefix = generate_new_prefix(X, prefix)
            if y_pred.ndim == 1:
                y_pred = Series(y_pred, index=X.index, name=prefix).to_frame()
            else:
                y_pred = DataFrame(y_pred, index=X.index)
                y_pred = y_pred.add_prefix(prefix + "_")
            y_pred.columns = y_pred.columns.str.replace(
                r"[^a-zA-Z0-9_]", "_", regex=True
            )

            # concatenate
            if self.concat:
                return concat([X, y_pred], axis=1)
            return y_pred
        else:
            # numpy
            if y_pred.ndim == 1:
                y_pred = y_pred.reshape(-1, 1)

            # concatenate
            if self.concat:
                return np.hstack([X, y_pred])
            return y_pred


class AppendPredictionToX(BaseEstimator, TransformerMixin, Generic[TEstimator]):
    """Append the prediction of the estimators to X."""

    estimators: Sequence[TEstimator]
    estimators_: Sequence[TEstimator]

    def __init__(
        self,
        estimators: Sequence[TEstimator] | TEstimator,
        *,
        concat: bool = True,
        n_jobs: int | None = -1,
    ) -> None:
        """Append the prediction of the estimators to X.

        If pandas DataFrame is given, the prediction is added as a new column
        with the name
        "y_pred_{estimator.__class__.__name__}_{i}_{estimator_index}"
        if the prediction is 1D, or
        "y_pred_{estimator.__class__.__name__}_{i}_{column_name}_{estimator_index}"
        if the prediction is 2D.

        Parameters
        ----------
        estimators : Sequence[TEstimator] | TEstimator
            The estimator(s) to be wrapped.
        concat : bool, optional
            Whether to concatenate the prediction to X,
            by default True
        n_jobs : int | None, optional
            The number of jobs to run in parallel, by default -1
        """
        if not isinstance(estimators, Sequence):
            estimators = [estimators]
        self.estimators = estimators
        self.concat = concat
        self.n_jobs = n_jobs

    def fit(self, X: Any, y: Any = None, **fit_params: Any) -> Self:
        """Fit the estimators."""
        estimators_ = Parallel(n_jobs=self.n_jobs)(
            delayed(AppendPredictionToXSingle(clone(estimator), concat=False).fit)(
                X, y, **fit_params
            )
            for estimator in self.estimators
        )
        if estimators_ is None:
            raise RuntimeError("Failed to fit estimators")
        self.estimators_ = estimators_
        return self

    def transform(self, X: TX, y: Any = None, **predict_params: Any) -> TX:
        """Append the prediction of the estimators to X.

        If pandas DataFrame is given, the prediction is added as a new column
        with the name
        "y_pred_{estimator.__class__.__name__}_{i}_{estimator_index}"
        if the prediction is 1D, or
        "y_pred_{estimator.__class__.__name__}_{i}_{column_name}_{estimator_index}"
        if the prediction is 2D."""
        check_is_fitted(self, "estimators_")
        transformed: list[TX] = [
            estimator_.transform(X, y, **predict_params).add_suffix(f"_{i}")
            for i, estimator_ in enumerate(self.estimators_)
        ]
        if isinstance(X, DataFrame):
            if self.concat:
                return concat([X, *transformed], axis=1)
            return concat(transformed, axis=1)
        else:
            if self.concat:
                return np.hstack([X, *transformed])
            return np.hstack(transformed)



from __future__ import annotations

import warnings
from typing import Any, Callable, Literal

from pandas import DataFrame, Index
from sklearn.base import BaseEstimator, TransformerMixin
from typing_extensions import Self

from .types import TXPandas


class ReindexMissingColumns(BaseEstimator, TransformerMixin):
    """Reindex X to match the columns of the training data to avoid errors."""

    def __init__(
        self,
        *,
        if_missing: Literal["warn", "raise"]
        | Callable[[Index[Any], Index[Any]], None] = "warn",
        reindex_kwargs: dict[
            Literal["method", "copy", "level", "fill_value", "limit", "tolerance"], Any
        ] = {},
    ) -> None:
        """Reindex X to match the columns of the training data to avoid errors.

        Parameters
        ----------
        if_missing : Literal['warn', 'raise'] | Callable[[Index[Any], Index[Any]], None], optional
            If callable, the first argument is the expected columns and the
            second argument is the actual columns, by default 'warn'
        reindex_kwargs : dict[Literal['method', 'copy', 'level', 'fill_value',
        'limit', 'tolerance'], Any], optional
            Keyword arguments to pass to reindex, by default {}
        """
        self.if_missing = if_missing
        self.reindex_kwargs = reindex_kwargs

    def fit(self, X: DataFrame, y: Any = None, **fit_params: Any) -> Self:
        self.feature_names_in_ = X.columns
        return self

    def transform(self, X: TXPandas, y: Any = None, **fit_params: Any) -> TXPandas:
        expected_columns = self.feature_names_in_
        actual_columns = X.columns
        if not expected_columns.equals(actual_columns):
            missing_columns = expected_columns.difference(actual_columns)
            if self.if_missing == "warn":
                warnings.warn(f"Missing columns: {missing_columns}")
            elif self.if_missing == "raise":
                raise ValueError(f"Missing columns: {missing_columns}")
            elif isinstance(self.if_missing, Callable):  # type: ignore
                self.if_missing(expected_columns, actual_columns)
            else:
                raise ValueError(f"Invalid value for if_missing: {self.if_missing}")

        return X.reindex(columns=self.feature_names_in_, **self.reindex_kwargs)



from typing import Any, Generic

from typing_extensions import Self

from .estimator_wrapper import EstimatorWrapperBase
from .types import TEstimator
from .utils import intersect_X_y


class IntersectXY(EstimatorWrapperBase[TEstimator], Generic[TEstimator]):
    """Estimator wrapper that intersects X and y indices before fitting."""

    def __init__(self, estimator: TEstimator) -> None:
        """Estimator wrapper that intersects X and y indices before fitting.
        Parameters
        ----------
        estimator : TEstimator
            The estimator to be wrapped.
        """
        super().__init__(estimator)

    def fit(self, X: Any, y: Any, **fit_params: Any) -> Self:
        X, y = intersect_X_y(X, y)
        self.estimator.fit(X, y, **fit_params)
        return self
